package net.knarcraft.blacksmith.formatting;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import static net.knarcraft.blacksmith.formatting.StringFormatter.replacePlaceholder;

/**
 * A utility for formatting a string specifying an amount of time
 */
public final class TimeFormatter {

    private static Map<Double, TranslatableMessage[]> timeUnits;
    private static List<Double> sortedUnits;

    private TimeFormatter() {

    }

    /**
     * Formats the given number of remaining seconds to the appropriate string
     *
     * @param exact   <p>Whether to display the exact number of seconds, or some unclear expression</p>
     * @param seconds <p>The number of seconds remaining</p>
     * @return <p>The time formatted correctly</p>
     */
    public static String formatTime(boolean exact, int seconds) {
        if (exact) {
            return getDurationString(seconds);
        } else {
            return formatUnclearTime(seconds);
        }
    }

    /**
     * Gets a time format using vague wording to specify a known interval
     *
     * @param seconds <p>The number of seconds to format</p>
     * @return <p>Text describing approximate time remaining</p>
     */
    private static String formatUnclearTime(int seconds) {
        List<TimeInterval> intervals = Arrays.stream(TimeInterval.values()).sorted().toList();
        for (TimeInterval interval : intervals) {
            if (seconds < interval.getIntervalMax()) {
                return getMessageFromInterval(interval);
            }
        }
        return TimeInterval.INTERVAL_MORE_THAN_5_MINUTES.getDefaultText();
    }

    /**
     * Gets the text string corresponding to the given time interval
     *
     * @param interval <p>The time interval to get text for</p>
     * @return <p>Text describing the time interval</p>
     */
    private static String getMessageFromInterval(TimeInterval interval) {
        String text = Translator.getTranslatedMessage(TranslatableMessage.valueOf(interval.name()));

        //Choose a random entry if a comma-separated list is provided
        if (text != null && text.contains(",")) {
            String[] parts = text.split(",");
            String randomPart = parts[new Random().nextInt(parts.length)];
            if (randomPart != null) {
                text = randomPart;
            }
        }

        //Use the set message, or use the default
        if (text != null && !text.trim().isEmpty()) {
            return text;
        } else {
            return interval.getDefaultText();
        }
    }

    /**
     * Gets the string used for displaying this sign's duration
     *
     * @return <p>The string used for displaying this sign's duration</p>
     */
    public static String getDurationString(int duration) {
        if (duration == 0) {
            return Translator.getTranslatedMessage(TranslatableMessage.UNIT_NOW);
        } else {
            if (sortedUnits == null) {
                initializeUnits();
            }
            for (Double unit : sortedUnits) {
                if (duration / unit >= 1) {
                    double units = round(duration / unit);
                    return formatDurationString(units, timeUnits.get(unit)[units == 1 ? 0 : 1],
                            (units * 10) % 10 == 0);
                }
            }
            return formatDurationString(duration, TranslatableMessage.UNIT_SECONDS, false);
        }
    }

    /**
     * Rounds a number to its last two digits
     *
     * @param number <p>The number to round</p>
     * @return <p>The rounded number</p>
     */
    private static double round(double number) {
        return Math.round(number * 100.0) / 100.0;
    }

    /**
     * Formats a duration string
     *
     * @param duration            <p>The duration to display</p>
     * @param translatableMessage <p>The time unit to display</p>
     * @param castToInt           <p>Whether to cast the duration to an int</p>
     * @return <p>The formatted duration string</p>
     */
    private static String formatDurationString(double duration, TranslatableMessage translatableMessage, boolean castToInt) {
        String durationFormat = Translator.getTranslatedMessage(TranslatableMessage.DURATION_FORMAT);
        durationFormat = replacePlaceholder(durationFormat, "{unit}",
                Translator.getTranslatedMessage(translatableMessage));
        return replacePlaceholder(durationFormat, "{time}", castToInt ? String.valueOf((int) duration) :
                String.valueOf(duration));
    }

    /**
     * Initializes the mapping of available time units for formatting permission sign duration
     */
    private static void initializeUnits() {
        double minute = 60;

        timeUnits = new HashMap<>();
        timeUnits.put(minute, new TranslatableMessage[]{TranslatableMessage.UNIT_MINUTE, TranslatableMessage.UNIT_MINUTES});
        timeUnits.put(1D, new TranslatableMessage[]{TranslatableMessage.UNIT_SECOND, TranslatableMessage.UNIT_SECONDS});

        sortedUnits = new ArrayList<>(timeUnits.keySet());
        Collections.sort(sortedUnits);
        Collections.reverse(sortedUnits);
    }

}
